#!/usr/bin/env python

'''
Server mode using Gevent, using tunnelling feature to give a remote shell.
'''

from __future__ import absolute_import

import os
import sys
import pty
import time
import subprocess
import threading
import logging
import fcntl
import socket

from sjrpc.server import RpcServer
from sjrpc.utils import RpcHandler, pass_connection
from sjrpc.core.protocols import TunnelProtocol

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
fmt = logging.Formatter('%(levelname)s %(name)s %(message)s')
handler = logging.StreamHandler()
handler.setFormatter(fmt)
logger.addHandler(handler)


class _FakePtySocket(object):

    '''
    A fake socket object wrapping a :class:`subprocess.Popen` object standard
    input/output.
    '''

    def __init__(self, fd):
        self._fd = fd

    def recv(self, size):
        return os.read(self._fd, size)

    def send(self, data):
        return os.write(self._fd, data)

    def fileno(self):
        return self._fd

    def setblocking(self, blocking):
        if not blocking:
            # Disable blocking mode on stdin:
            current_flags = fcntl.fcntl(self._fd, fcntl.F_GETFL)
            fcntl.fcntl(self._fd, fcntl.F_SETFL, current_flags | os.O_NONBLOCK)
            pass
        else:
            raise NotImplementedError('Not implemented')


class _FakeDevZeroSocket(object):

    def __init__(self):
        self._devzero = open('/dev/zero', 'rb')

    def recv(self, size):
        return self._devzero.read(size)

    def send(self, data):
        pass # Cannot write on /dev/zero

    def fileno(self):
        return self._devzero.fileno()

    def setblocking(self, blocking):
        if not blocking:
            # Disable blocking mode on devzero:
            current_flags = fcntl.fcntl(self.fileno(), fcntl.F_GETFL)
            fcntl.fcntl(self.fileno(), fcntl.F_SETFL, current_flags | os.O_NONBLOCK)
            pass
        else:
            raise NotImplementedError('Not implemented')


class RemoteShellHandler(RpcHandler):

    @pass_connection
    def launch_remote_console(self, connection, shell='/bin/bash'):

        master, slave = pty.openpty()
        endpoint = _FakePtySocket(master)
        # Create the tunnel on the label 1
        connection.register_protocol(1, TunnelProtocol, endpoint=endpoint)
        # Launch the shell process:
        self.pshell = subprocess.Popen(shell, stdout=slave, bufsize=0,
                                  stdin=slave, cwd='/',
                                  stderr=subprocess.STDOUT)
        return 1 # Return the tunnel id.

    #@pass_connection
    #def iperf(self, connection):
    #    devzero = _FakeDevZeroSocket()
    #    tun = connection.register_protocol(2, TunnelProtocol, endpoint=devzero)
    #    return 2

    @pass_connection
    def wait_for_pshell(self, conn):
        #conn.get_protocol(1)._send_get(1024 * 1024)
        return self.pshell.wait()

    def hup(self, *args, **kwargs):
        self.pshell.kill()

# Get arguments from the command line:
if len(sys.argv) < 3:
    print 'Usage: %s <listening address> <listening port>' % sys.argv[0]
    sys.exit(2)
address = sys.argv[1]
port = int(sys.argv[2])

# Create the server instance:
import pyev
loop = pyev.default_loop(flags=~ pyev.EVBACKEND_EPOLL)
rpcserver = RpcServer.from_addr(address, port, loop=loop, conn_kw=dict(handler=RemoteShellHandler(), on_disconnect='hup'))
rpcserver.run()

